/*
  ==============================================================================

   This file is part of the Plugin Consultant - VST API Logger
   Copyright 2005 by Ofer Shem Tov.

  ------------------------------------------------------------------------------

   Plugin Consultant can be redistributed and/or modified under the terms of the
   GNU General Public License, as published by the Free Software Foundation;
   either version 2 of the License, or (at your option) any later version.

   Plugin Consultant is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Plugin Consultant; if not, visit www.gnu.org/licenses or write to the
   Free Software Foundation, Inc., 59 Temple Place, Suite 330,
   Boston, MA 02111-1307 USA

  ==============================================================================
*/

#ifdef _MSC_VER
#pragma warning (disable : 4996)
#endif

#include <AEffectx.h>
#include <juce.h>
#include "PCLogger.h"

#ifdef JUCE_WIN32
  #undef Rectangle
  #define Array tempArray
  #include <windows.h>
  #undef Array
#endif

// true if logger initialized already
static bool initialized = false;

enum {

	pcLogMax = 1024
};

//-----------------------------------------------------------------------------------------
// VST C functions
enum vstFunction {
	vstDispatcher,
	vstAudioMaster,
	vstProcess,
	vstProcessReplacing,
	vstSetParameter,
	vstGetParameter,
	vstNumFunctions
};
//-----------------------------------------------------------------------------------------
static char* vstFunctionNames[] =
{
	"dispatcher",
	"audioMaster",
	"process",
	"processReplacing",
	"setParameter",
	"getParameter",
};

//-----------------------------------------------------------------------------------------
// possible dispatcher ptr types
enum PtrType {

	ptrNone,
	ptrChar,
	ptrERect,
	ptrPtrVoid,
	ptrVoid,
	ptrEvents,
	ptrPinProperties,
	ptrParameterProperties,
	ptrTimeInfo,
	ptrWindow,
	ptrFileSelect,
	ptrAudioFile,
	ptrOfflineTask,
	ptrVariableIo,
	ptrSpeakerArrangement,
	ptrMidiProgramName,
	ptrMidiProgramCategory,
	ptrMidiKeyName,
	ptrPatchChunkInfo,
};

//-----------------------------------------------------------------------------------------
// information about a VST opcode
struct opCodeDebugInfo {

	char*	opCodeName;	// opcode name
	char*	longName;	// long return value name or NULL if no return value
	char*	indexName;	// index parameter name or NULL if no index parameter
	char*	valueName;	// value parameter name or NULL if no value parameter
	bool	ptrReturn;	// true if ptr is a return value
	PtrType	ptrType;	// type of data ptr is pointing to, ptrNone if no ptr parameter
	char*	ptrName;	// name of ptr parameter or NULL if no ptr parameter
	char*	optName;	// float parameter name or NULL if no float parameter
};

//-----------------------------------------------------------------------------------------
static struct opCodeDebugInfo effOpCodes[] =
{
	// name							long		index		value		ptr-return	ptr-type		ptr			float
	{ "Open",						NULL,		NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "Close",						NULL,		NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "SetProgram",					NULL,		NULL,		"program",	false,		ptrNone,		NULL,		NULL },
	{ "GetProgram",					"program",	NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "SetProgramName",				NULL,		NULL,		NULL,		false,		ptrChar,		"name",		NULL },
	{ "GetProgramName",				NULL,		NULL,		NULL,		true,		ptrChar,		"name",		NULL },
	{ "GetParamLabel",				NULL,		"parameter",NULL,		true,		ptrChar,		"label",	NULL },
	{ "GetParamDisplay",			NULL,		"parameter",NULL,		true,		ptrChar,		"value",	NULL },
	{ "GetParamName",				NULL,		"parameter",NULL,		true,		ptrChar,		"name",		NULL },
	{ "GetVu",						"level",	NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "SetSampleRate",				NULL,		NULL,		NULL,		false,		ptrNone,		NULL,		"sampleRate" },
	{ "SetBlockSize",				NULL,		NULL,		"blockSize",false,		ptrNone,		NULL,		NULL },
	{ "MainsChanged",				NULL,		NULL,		"state",	false,		ptrNone,		NULL,		NULL },
	{ "EditGetRect",				"long",		NULL,		NULL,		false,		ptrERect,		NULL,		NULL },
	{ "EditOpen",					"long",		NULL,		NULL,		false,		ptrVoid,		"window",	NULL },
	{ "EditClose",					NULL,		NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "EditDraw",					NULL,		NULL,		NULL,		false,		ptrERect,		"rect",		NULL },
	{ "EditMouse",					"long",		"x",		"y",		false,		ptrNone,		NULL,		NULL },
	{ "EditKey",					"long",		NULL,		"keyCode",	false,		ptrNone,		NULL,		NULL },
	{ "EditIdle",					NULL,		NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "EditTop",					NULL,		NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "EditSleep",					NULL,		NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "Identify",					"CCONST",	NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "GetChunk",					"dataSize",	"isPreset",	NULL,		false,		ptrPtrVoid,		"data",		NULL },
	{ "SetChunk",					"long",		"isPreset",	"byteSize",	false,		ptrVoid,		"data",		NULL },
	{ "ProcessEvents",				"long",		NULL,		NULL,		false,		ptrEvents,		"events",	NULL },
	{ "CanBeAutomated",				"bool",		"parameter",NULL,		false,		ptrNone,		NULL,		NULL },
	{ "String2Parameter",			"bool",		"parameter",NULL,		false,		ptrChar,		"text",		NULL },
	{ "GetNumProgramCategories",	"categories",NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "GetProgramNameIndexed",		"bool",		"index",	"category",	true,		ptrChar,		"text",		NULL },
	{ "CopyProgram",				"bool",		"destination",NULL,		false,		ptrNone,		NULL,		NULL },
	{ "ConnectInput",				NULL,		"index",	"state",	false,		ptrNone,		NULL,		NULL },
	{ "ConnectOutput",				NULL,		"index",	"state",	false,		ptrNone,		NULL,		NULL },
	{ "GetInputProperties",			"bool",		"index",	NULL,		false,		ptrPinProperties,"properties",NULL },
	{ "GetOutputProperties",		"bool",		"index",	NULL,		false,		ptrPinProperties,"properties",NULL },
	{ "GetPlugCategory",			"category",	NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "GetCurrentPosition",			"long",		NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "GetDestinationBuffer",		"float*",	NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "OfflineNotify",				"bool",		"start",	"numAudioFiles",false,	ptrAudioFile,	"audioFiles",NULL },
	{ "OfflinePrepare",				"bool",		NULL,		"count",	false,		ptrOfflineTask,	"offline",	NULL },
	{ "OfflineRun",					"bool",		NULL,		"count",	false,		ptrOfflineTask,	"offline",	NULL },
	{ "ProcessVarIo",				"bool",		NULL,		NULL,		false,		ptrVariableIo,	"varIo",	NULL },
	{ "SetSpeakerArrangement",		"bool",		NULL,		"pluginInput",false,	ptrSpeakerArrangement,"pluginOutput",NULL },
	{ "SetBlockSizeAndSampleRate",	NULL,		NULL,		"blockSize",false,		ptrNone,		NULL,		"sampleRate" },
	{ "SetBypass",					"bool",		NULL,		"onOff",	false,		ptrNone,		NULL,		NULL },
	{ "GetEffectName",				"bool",		NULL,		NULL,		true,		ptrChar,		"name",		NULL },
	{ "GetErrorText",				"bool",		NULL,		NULL,		true,		ptrChar,		"text",		NULL },
	{ "GetVendorString",			"bool",		NULL,		NULL,		true,		ptrChar,		"text",		NULL },
	{ "GetProductString",			"bool",		NULL,		NULL,		true,		ptrChar,		"text",		NULL },
	{ "GetVendorVersion",			"version",	NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "VendorSpecific",				"long",		"lArg1",	"lArg2",	false,		ptrVoid,		"prtArg",	"floatArg" },
	{ "CanDo",						"long",		NULL,		NULL,		false,		ptrChar,		"text",		NULL },
	{ "GetTailSize",				"long",		NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "Idle",						"long",		NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "GetIcon",					"void*",	NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "SetViewPosition",			"bool",		"x",		"y",		false,		ptrNone,		NULL,		NULL },
	{ "GetParameterProperties",		"bool",		"index",	NULL,		false,		ptrParameterProperties, "properties",NULL },
	{ "KeysRequired",				"bool",		NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "GetVstVersion",				"version",	NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "EditKeyDown",				"long",		"character","virt",		false,		ptrNone,		NULL,		"modifier" },
	{ "EditKeyUp",					"long",		"character","virt",		false,		ptrNone,		NULL,		"modifier" },
	{ "SetEditKnobMode",			"long",		NULL,		"value",	false,		ptrNone,		NULL,		NULL },
	{ "GetMidiProgramName",			"programIndexes","MidiChannel",NULL,false,		ptrMidiProgramName,"program",NULL },
	{ "GetCurrentMidiProgram",		"programIndex",	NULL,	NULL,		false,		ptrMidiProgramName,"program",NULL },
	{ "GetMidiProgramCategory",		"categoryIndexes",NULL,	NULL,		false,		ptrMidiProgramCategory,"category",NULL },
	{ "HasMidiProgramsChanged",		"bool",		"MidiChannel",NULL,		false,		ptrNone,		NULL,		NULL },
	{ "GetMidiKeyName",				"bool",		NULL,		NULL,		false,		ptrMidiKeyName,	"key",		NULL },
	{ "BeginSetProgram",			"bool",		NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "EndSetProgram",				"bool",		NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "GetSpeakerArrangement",		"bool",		NULL,		"pluginInput",false,	ptrSpeakerArrangement,"pluginOutput",NULL },
	{ "ShellGetNextPlugin",			"uniqueID",	NULL,		NULL,		true,		ptrChar,		"name",		NULL },
	{ "StartProcess",				"long",		NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "StopProcess",				"long",		NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "SetTotalSampleToProcess",	"long",		NULL,		"value",	false,		ptrNone,		NULL,		NULL },
	{ "SetPanLaw",					"bool",		NULL,		"type",		false,		ptrNone,		NULL,		"gain" },
	{ "BeginLoadBank",				"long",		NULL,		NULL,		false,		ptrPatchChunkInfo,"patchChunkInfo",NULL },
	{ "BeginLoadProgram",			"long",		NULL,		NULL,		false,		ptrPatchChunkInfo,"patchChunkInfo",NULL },
};

//-----------------------------------------------------------------------------------------
static struct opCodeDebugInfo audioMasterOpCodes[] =
{
	// name								long		index		value		ptr-return	ptr-type		ptr			float
	{ "Automate ",						NULL,		"index",	NULL,		false,		ptrNone,		NULL,		"value" },
	{ "Version",						"version",	NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "CurrentId",						"Unique ID",NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "Idle",							NULL,		NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "PinConnected",					"connected","pin index","isOutput",	false,		ptrNone,		NULL,		NULL },
	{ "ThereIsNoSuchOpCode",			NULL,		NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "WantMidi",						NULL,		NULL,		"filter",	false,		ptrNone,		NULL,		NULL },
	{ "GetTime",						"TimeInfo",	NULL,		"filter",	false,		ptrNone,		NULL,		NULL },
	{ "ProcessEvents",					"bool",		NULL,		NULL,		false,		ptrEvents,		"events",	NULL },
	{ "SetTime",						NULL,		NULL,		"filter",	false,		ptrTimeInfo,	"TimeInfo", NULL },
	{ "TempoAt",						"tempo",	NULL,		"pos",		false,		ptrNone,		NULL,		NULL },
	{ "GetNumAutomatableParameters",	"long",		NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "GetParameterQuantization",		"long",		NULL,		"index",	false,		ptrNone,		NULL,		NULL },
	{ "IOChanged",						"bool",		NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "NeedIdle",						"bool",		NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "SizeWindow",						"bool",		"width",	"height",	false,		ptrNone,		NULL,		NULL },
	{ "GetSampleRate",					"sampleRate",NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "GetBlockSize",					"blockSize",NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "GetInputLatency",				"latency",	NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "GetOutputLatency",				"latency",	NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "GetPreviousPlug",				"effect",	NULL,		"inpuPin",	false,		ptrNone,		NULL,		NULL },
	{ "GetNextPlug",					"effect",	NULL,		"outputPin",false,		ptrNone,		NULL,		NULL },
	{ "WillReplaceOrAccumulate",		"long",		NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "GetCurrentProcessLevel",			"long",		NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "GetAutomationState",				"long",		NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "OfflineStart",					"bool",		"numNewAudioFiles","numAudioFiles",	false,ptrAudioFile,"audioFiles",NULL },
	{ "OfflineRead",					"bool",		"readSource","option",	false,		ptrOfflineTask,	"offline",	NULL },
	{ "OfflineWrite",					"bool",		NULL,		"option",	false,		ptrOfflineTask,	"offline",	NULL },
	{ "OfflineGetCurrentPass",			"long",		NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "OfflineGetCurrentMetaPass",		"long",		NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "SetOutputSampleRate",			NULL,		NULL,		NULL,		false,		ptrNone,		NULL,		"sampleRate" },
	{ "GetSpeakerArrangement",			NULL,		NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "GetVendorString",				"bool",		NULL,		NULL,		true,		ptrChar,		"text",		NULL },
	{ "GetProductString",				"bool",		NULL,		NULL,		true,		ptrChar,		"text",		NULL },
	{ "GetVendorVersion",				"version",	NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "VendorSpecific",					"long",		"lArg1",	"lArg2",	false,		ptrVoid,		"prtArg",	"floatArg" },
	{ "SetIcon",						NULL,		NULL,		NULL,		false,		ptrVoid,		"icon",		NULL },
	{ "CanDo",							"long",		NULL,		NULL,		false,		ptrChar,		"text",		NULL },
	{ "GetLanguage",					"language",	NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "OpenWindow",						"bool",		NULL,		NULL,		false,		ptrWindow,		"window",	NULL },
	{ "CloseWindow",					"bool",		NULL,		NULL,		false,		ptrWindow,		"window",	NULL },
	{ "GetDirectory",					"plugDir",	NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "UpdateDisplay",					"bool",		NULL,		NULL,		false,		ptrNone,		NULL,		NULL },
	{ "BeginEdit",						"bool",		"index",	NULL,		false,		ptrNone,		NULL,		NULL },
	{ "EndEdit",						"bool",		"index",	NULL,		false,		ptrNone,		NULL,		NULL },
	{ "OpenFileSelector",				"bool",		NULL,		NULL,		false,		ptrFileSelect,	"VstFileSelect*",NULL },
	{ "CloseFileSelector",				"bool",		NULL,		NULL,		false,		ptrFileSelect,	"VstFileSelect*",NULL },
	{ "EditFile",						NULL,		NULL,		NULL,		false,		ptrChar,		"xml",		NULL },
	{ "GetChunkFile",					"bool",		NULL,		NULL,		false,		ptrVoid,		"nativePath",NULL },
	{ "GetInputSpeakerArrangement",		"speakerArrangement",NULL,NULL,		false,		ptrNone,		NULL,		NULL },
};

//-----------------------------------------------------------------------------------------

struct DynamicDebugInfo {

	long count;		// how many time the opcode/function was called
	bool enabled;	// log is enabled for this opcode/function
};

enum {
	numEffOpCodes = effNumV2_3Opcodes,									// eff op codes
	numAudioMasterOpCodes = audioMasterGetInputSpeakerArrangement + 1,	// audioMaster op codes
};

// arrays of all dynamic debug info for all VST functions and opCodes
static struct DynamicDebugInfo vstFunctionsLogInfo[vstNumFunctions];
static struct DynamicDebugInfo vstDispatcherLogInfo[numEffOpCodes ];
static struct DynamicDebugInfo vstAudioMasterLogInfo[numAudioMasterOpCodes];

//-----------------------------------------------------------------------------------------

struct vstOpCodeGroupInfo {

	int numOpCodes;					// number of op codes in group
	char* name;						// opCodes group name
	DynamicDebugInfo* debugInfo;	// group debug info
};

// Groups of OpCodes supported by the logger
static vstOpCodeGroupInfo OpCodeGroupsInfo[PCLogger::vstNumOpCodeGroups] = {

	{ vstNumFunctions, "vstAEffectFunctions", vstFunctionsLogInfo },
	{ numEffOpCodes , "vstDispatcherOpCodes", vstDispatcherLogInfo },
	{ numAudioMasterOpCodes, "vstAudioMasterOpCodes", vstAudioMasterLogInfo }
};

//-----------------------------------------------------------------------------------------
void initLogInfo( DynamicDebugInfo* info, int numInfos )
{
	for ( int i = 0; i < numInfos; i++ )
	{
		info->count = 0;
		info->enabled = true;
		info++;
	};
}

void initLogCounters( DynamicDebugInfo* info, int numInfos )
{
	for ( int i = 0; i < numInfos; i++ )
	{
		info->count = 0;
		info++;
	};
}

inline DynamicDebugInfo* getLogInfo( PCLogger::vstOpCodeGroup group, long opCode )
{
	if ( group < 0 || group >= PCLogger::vstNumOpCodeGroups )
		return NULL;

	return &OpCodeGroupsInfo[group].debugInfo[opCode];
}

//-----------------------------------------------------------------------------------------
void PCLogger::logMessage ( const String& message )
{
	log( message );
}

//-----------------------------------------------------------------------------------------
void PCLogger::log( const char* message )
{
	// for now start with a simple log...
	char buf[pcLogMax];
	sprintf( buf, "(%d) %s:%s", Thread::getCurrentThreadId(), (const char*)plugName, (const char*)message );
	outputDebugString ( buf );
}

//-----------------------------------------------------------------------------------------
void PCLogger::init( const String& name )
{
	logOn = true;

	if ( initialized )
		return;

	logParams = true;
	logRetParams = true;

	plugName = name;

	// init dynamic debug info data
	initLogInfo( vstFunctionsLogInfo, vstNumFunctions );
	initLogInfo( vstDispatcherLogInfo, numEffOpCodes );
	initLogInfo( vstAudioMasterLogInfo, numAudioMasterOpCodes );

	Logger::setCurrentLogger( this );
	initialized = true;
}

//-----------------------------------------------------------------------------------------
#ifdef JUCE_WIN32
void PCLogger::logDllMain( long dwReason )
{
	if ( !isOn() )
		return;

	if ( dwReason == DLL_PROCESS_ATTACH )
		log( "DLL_PROCESS_ATTACH" );
	else if ( dwReason == DLL_PROCESS_DETACH )
		log( "DLL_PROCESS_DETACH" );
	else if ( dwReason == DLL_THREAD_ATTACH )
		log( "DLL_THREAD_ATTACH" );
	else if ( dwReason == DLL_THREAD_DETACH )
		log( "DLL_THREAD_DETACH" );
	else
		log( "DllMain" );
}
#endif // JUCE_WIN32

//-----------------------------------------------------------------------------------------
void PCLogger::logVstDispatcher(long opCode, long index, long value, void *ptr, float opt)
{
	// if opCode is newer than VST SDK 2.3 then ignore
	if ( opCode < effOpen || opCode >= numEffOpCodes )
		return;

	// increment counter for dispatcher and opCode
	vstFunctionsLogInfo[vstDispatcher].count++;
	vstDispatcherLogInfo[opCode].count++;

	// log only if function and opCode logging are enabled
	if ( !isOn() || !vstFunctionsLogInfo[vstDispatcher].enabled || !vstDispatcherLogInfo[opCode].enabled )
		return;

	char buf[pcLogMax];
	sprintf( buf, "eff%s ", effOpCodes[opCode].opCodeName );

	// log parameters if enabled
	if ( isLogParameters() )
	{
		int nextChar = strlen( buf );

		// get parameters name and values if relevant for this opCode
		if ( effOpCodes[opCode].indexName != NULL )
				nextChar += sprintf( &buf[nextChar], "%s=%d ", effOpCodes[opCode].indexName, index );

		if ( effOpCodes[opCode].valueName != NULL )
			nextChar += sprintf( &buf[nextChar], "%s=%d ", effOpCodes[opCode].valueName, value );

		// even if there is a ptr value don't log it if its a return value
		if ( effOpCodes[opCode].ptrName != NULL && !effOpCodes[opCode].ptrReturn )
		{
			if ( effOpCodes[opCode].ptrType == ptrChar )
				nextChar += sprintf( &buf[nextChar], "%s=%s ", effOpCodes[opCode].ptrName, (char*) ptr );
			else
				nextChar += sprintf( &buf[nextChar], "%s=%p ", effOpCodes[opCode].ptrName, ptr );
		}

		if ( effOpCodes[opCode].optName != NULL )
			nextChar += sprintf( &buf[nextChar], "%s=%f ", effOpCodes[opCode].optName, opt );
	}

	log( buf );
}

//-----------------------------------------------------------------------------------------
void	PCLogger::logVstDispatcherReturn( long opCode, void *ptr, long result )
{
	// do nothing if not logging return parameters
	if ( !isLogReturnParameters() )
		return;

	// if opCode is newer than VST SDK 2.3 then ignore
	if ( opCode < effOpen || opCode >= numEffOpCodes )
		return;

	// log only if function and opCode logging are enabled
	if ( !isOn() || !vstFunctionsLogInfo[vstDispatcher].enabled || !vstDispatcherLogInfo[opCode].enabled )
		return;

	// do nothing if this opCode don't have any return values
	if ( effOpCodes[opCode].longName == NULL && !effOpCodes[opCode].ptrReturn )
		return;

	char buf[pcLogMax];
	sprintf( buf, "eff%s returned ", effOpCodes[opCode].opCodeName );

	int nextChar = strlen( buf );

	// see if a long result is relevant for this opCode
	if ( effOpCodes[opCode].longName != NULL )
		nextChar += sprintf( &buf[nextChar], "%s=%d ", effOpCodes[opCode].longName, result );

	// log only if its a ptr return value and a name for the pointer paramter
	if ( effOpCodes[opCode].ptrName != NULL && effOpCodes[opCode].ptrReturn )
	{
		if ( effOpCodes[opCode].ptrType == ptrChar )
			nextChar += sprintf( &buf[nextChar], "%s=%s", effOpCodes[opCode].ptrName, (char*) ptr );
		else
			nextChar += sprintf( &buf[nextChar], "%s=%p", effOpCodes[opCode].ptrName, ptr );
	}

	log( buf );
}

//-----------------------------------------------------------------------------------------
void PCLogger::logVstAudioMaster( long opCode, long index, long value, void *ptr, float opt )
{
	// if opCode is newer than VST SDK 2.3 then ignore it
	if ( opCode < audioMasterAutomate  || opCode >= numAudioMasterOpCodes )
		return;

	// increment counter for audioMaster and opCode
	vstFunctionsLogInfo[vstAudioMaster].count++;
	vstAudioMasterLogInfo[opCode].count++;

	// log only if function and opCode logging are enabled
	if ( !isOn() || !vstFunctionsLogInfo[vstAudioMaster].enabled || !vstAudioMasterLogInfo[opCode].enabled )
		return;

	char buf[pcLogMax];
	sprintf( buf, "audioMaster%s ", audioMasterOpCodes[opCode].opCodeName );

	// log parameters if enabled
	if ( isLogParameters() )
	{
		int nextChar = strlen( buf );

		// get parameters name and values if relevant for this opCode
		if ( audioMasterOpCodes[opCode].indexName != NULL )
				nextChar += sprintf( &buf[nextChar], "%s=%d ", audioMasterOpCodes[opCode].indexName, index );

		if ( audioMasterOpCodes[opCode].valueName != NULL )
			nextChar += sprintf( &buf[nextChar], "%s=%d ", audioMasterOpCodes[opCode].valueName, value );

		// even if there is a ptr value don't log it if its a return value
		if ( audioMasterOpCodes[opCode].ptrName != NULL && !audioMasterOpCodes[opCode].ptrReturn )
		{
			if ( audioMasterOpCodes[opCode].ptrType == ptrChar )
				nextChar += sprintf( &buf[nextChar], "%s=%s ", audioMasterOpCodes[opCode].ptrName, (char*) ptr );
			else
				nextChar += sprintf( &buf[nextChar], "%s=%p ", audioMasterOpCodes[opCode].ptrName, ptr );
		}

		if ( audioMasterOpCodes[opCode].optName != NULL )
			nextChar += sprintf( &buf[nextChar], "%s=%f ", audioMasterOpCodes[opCode].optName, opt );
	}

	log( buf );
}

//-----------------------------------------------------------------------------------------
void PCLogger::logVstAudioMasterReturn( long opCode, void *ptr, long result )
{
	// do nothing if not logging return parameters
	if ( !isLogReturnParameters() )
		return;

	// if opCode is newer than VST SDK 2.3 then ignore it
	if ( opCode < audioMasterAutomate  || opCode >= numAudioMasterOpCodes )
		return;

	// log only if function and opCode logging are enabled
	if ( !isOn() || !vstFunctionsLogInfo[vstAudioMaster].enabled || !vstAudioMasterLogInfo[opCode].enabled )
		return;

	// do nothing if this opCode don't have any return values
	if ( audioMasterOpCodes[opCode].longName == NULL && !audioMasterOpCodes[opCode].ptrReturn )
		return;

	char buf[pcLogMax];
	sprintf( buf, "audioMaster%s returned ", audioMasterOpCodes[opCode].opCodeName );

	int nextChar = strlen( buf );

	// see if a long result is relevant for this opCode
	if ( audioMasterOpCodes[opCode].longName != NULL )
		nextChar += sprintf( &buf[nextChar], "%s=%d ", audioMasterOpCodes[opCode].longName, result );

	// log only if its a ptr return value and a name for the pointer paramter
	if ( audioMasterOpCodes[opCode].ptrName != NULL && audioMasterOpCodes[opCode].ptrReturn )
	{
		if ( audioMasterOpCodes[opCode].ptrType == ptrChar )
			nextChar += sprintf( &buf[nextChar], "%s=%s", audioMasterOpCodes[opCode].ptrName, (char*) ptr );
		else
			nextChar += sprintf( &buf[nextChar], "%s=%p", audioMasterOpCodes[opCode].ptrName, ptr );
	}

	log( buf );
}

//-----------------------------------------------------------------------------------------
void PCLogger::logVstProcess( float **inputs, float **outputs, long sampleframes, bool replacing )
{
	int func;

	if ( replacing )
		func = vstProcessReplacing;
	else
		func = vstProcess;

	// increment counter for process/ProcessReplacing
	vstFunctionsLogInfo[func].count++;
	
	// log only if function is enabled
	if ( !isOn() || !vstFunctionsLogInfo[func].enabled )
		return;

	char buf[pcLogMax];

	if ( replacing )
		strcpy( buf, "processReplacing" );
	else
		strcpy( buf, "process" );

	// log parameters if enabled
	if ( isLogParameters() )
		sprintf( &buf[strlen( buf )], "(inputs:%p, outputs:%p, sampleFrames:%d)", inputs, outputs, sampleframes);
	
	log( buf );
}

//-----------------------------------------------------------------------------------------
void PCLogger::logVstSetParameter( long index, float parameter )
{
	// increment counter for set parameter
	vstFunctionsLogInfo[vstSetParameter].count++;

	// log only if function is enabled
	if ( !isOn() || !vstFunctionsLogInfo[vstSetParameter].enabled )
		return;

	char buf[pcLogMax];
	strcpy( buf, "setParameter" );

	// log parameters if enabled
	if ( isLogParameters() )
		sprintf( &buf[strlen( buf )], "(index: %d, parameter: %f)", index, parameter );

	log( buf );
}

//-----------------------------------------------------------------------------------------
void PCLogger::logVstGetParameter( long index )
{
	// increment counter for get parameter
	vstFunctionsLogInfo[vstGetParameter].count++;

	// log only if function is enabled
	if ( !isOn() || !vstFunctionsLogInfo[vstGetParameter].enabled )
		return;

	char buf[pcLogMax];
	strcpy( buf, "getParameter" );

	// log parameters if enabled
	if ( isLogParameters() )
		sprintf( &buf[strlen( buf )], "(index: %d)", index );

	log( buf );
}

//-----------------------------------------------------------------------------------------
void PCLogger::logVstGetParameterReturn( float parameter )
{
	// do nothing if not logging return parameters
	if ( !isLogReturnParameters() )
		return;

	// log only if function is enabled
	if ( !isOn() || !vstFunctionsLogInfo[vstGetParameter].enabled )
		return;

	char buf[pcLogMax];
	sprintf( buf, "getParameter returned %f", parameter );
	log( buf );
}

//-----------------------------------------------------------------------------------------
const String PCLogger::getVstOpCodeName( vstOpCodeGroup group, long opCode )
{
	String retName;

	switch ( group )
	{
		case vstFunctions:
			retName = vstFunctionNames[opCode];
			break;
		case vstDispatcherOpCodes:
			retName = effOpCodes[opCode].opCodeName;
			break;
		case vstAudioMasterOpCodes:
			retName = audioMasterOpCodes[opCode].opCodeName;
			break;
		default:
			retName = "unknown";
			break;
	}

	return retName;
}

//-----------------------------------------------------------------------------------------
int PCLogger::getVstNumOpCodes( vstOpCodeGroup group )
{
	if ( group < 0 || group >= PCLogger::vstNumOpCodeGroups )
		return 0;

	return OpCodeGroupsInfo[group].numOpCodes;
}

//-----------------------------------------------------------------------------------------
long PCLogger::getVstOpCodeCount( vstOpCodeGroup group, long opCode )
{
	return getLogInfo( group, opCode)->count;
}

//-----------------------------------------------------------------------------------------
bool PCLogger::isVstOpCodeEnabled( vstOpCodeGroup group, long opCode )
{
	return getLogInfo( group, opCode)->enabled;
}

//-----------------------------------------------------------------------------------------
void PCLogger::setVstOpCodeEnabled( bool enabled, vstOpCodeGroup group, long opCode )
{
	getLogInfo( group, opCode)->enabled = enabled;
}

//-----------------------------------------------------------------------------------------
void PCLogger::setVstOpCodesEnabled( bool enabled )
{
	for ( int group = 0; group < vstNumOpCodeGroups; group++ )
		for ( long opCode = 0; opCode < OpCodeGroupsInfo[group].numOpCodes; opCode++ )
			OpCodeGroupsInfo[group].debugInfo[opCode].enabled = enabled;
}

//-----------------------------------------------------------------------------------------
void PCLogger::toggleVstOpCodesEnabled()
{
	for ( int group = 0; group < vstNumOpCodeGroups; group++ )
		for ( long opCode = 0; opCode < OpCodeGroupsInfo[group].numOpCodes; opCode++ )
			OpCodeGroupsInfo[group].debugInfo[opCode].enabled = !OpCodeGroupsInfo[group].debugInfo[opCode].enabled;
}

//-----------------------------------------------------------------------------------------
void PCLogger::resetVstOpCodeCounters()
{
	initLogCounters( vstFunctionsLogInfo, vstNumFunctions );
	initLogCounters( vstDispatcherLogInfo, numEffOpCodes );
	initLogCounters( vstAudioMasterLogInfo, numAudioMasterOpCodes );
}

//-----------------------------------------------------------------------------------------
void PCLogger::loadState( const XmlElement& state )
{
	// load opcode preferences
	for ( int group = 0; group < vstNumOpCodeGroups; group++ )
	{
		// search for group element
		XmlElement* groupElement = state.getChildByName( OpCodeGroupsInfo[group].name );

		if ( groupElement != NULL )
		{
			// iterate opCode elements, identify with index and get opCode state
			XmlElement* opCodeElement = groupElement->getFirstChildElement();

			while ( opCodeElement != NULL )
			{
				if ( opCodeElement->hasTagName ( "opCode" ) )
				{
					long opCode = opCodeElement->getIntAttribute( "index", -1 );

					if ( opCode >= 0 && opCode < OpCodeGroupsInfo[group].numOpCodes )
					{
						int enabled = opCodeElement->getIntAttribute( "enabled", -1 );
						if ( enabled >= 0 )
							setVstOpCodeEnabled( enabled ? true : false, (PCLogger::vstOpCodeGroup)group, opCode );
					}
				}
				opCodeElement = opCodeElement->getNextElement();
            }

        }
	}

	// load global preferences
	XmlElement* paramsElement = state.getChildByName( "Parameters" );

	if ( paramsElement )
	{
		int logParams = paramsElement->getIntAttribute( "logParameters", -1 );
		if ( logParams >= 0 )
			setLogParameters( logParams ? true : false );

		int logReturnParams = paramsElement->getIntAttribute( "logReturnParameters", -1 );
		if ( logReturnParams >= 0 )
			setLogReturnParameters( logReturnParams ? true : false );
	}
}

//-----------------------------------------------------------------------------------------
void PCLogger::saveState( XmlElement& state )
{
	// clean everything
	state.deleteAllChildElements();

	// save opcode preferences
	for ( int group = 0; group < vstNumOpCodeGroups; group++ )
	{
		// create child element for each group
		XmlElement* groupElement = new XmlElement( OpCodeGroupsInfo[group].name );
		state.addChildElement( groupElement );

		// write OpCodes index, name and state
		for ( long opCode = 0; opCode < OpCodeGroupsInfo[group].numOpCodes; opCode++ )
		{
			XmlElement* opCodeElement = new XmlElement( "opCode" );
			opCodeElement->setAttribute( "index", opCode );
			opCodeElement->setAttribute( "name", getVstOpCodeName( (PCLogger::vstOpCodeGroup)group, opCode ) );
			opCodeElement->setAttribute( "enabled", isVstOpCodeEnabled( (PCLogger::vstOpCodeGroup)group, opCode ) );
			groupElement->addChildElement( opCodeElement );
		}
	}

	// save global preferences
	XmlElement* paramsElement = new XmlElement( "Parameters" );
	paramsElement->setAttribute( "logParameters", isLogParameters() );
	paramsElement->setAttribute( "logReturnParameters", isLogReturnParameters() );
	state.addChildElement( paramsElement );
}

