/*
  ==============================================================================

   This file is part of the Plugin Consultant - VST API Logger
   Copyright 2005 by Ofer Shem Tov.

  ------------------------------------------------------------------------------

   Plugin Consultant can be redistributed and/or modified under the terms of the
   GNU General Public License, as published by the Free Software Foundation;
   either version 2 of the License, or (at your option) any later version.

   Plugin Consultant is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Plugin Consultant; if not, visit www.gnu.org/licenses or write to the
   Free Software Foundation, Inc., 59 Temple Place, Suite 330,
   Boston, MA 02111-1307 USA

  ==============================================================================
*/

class PCLogger : public Logger {

public:

	//------------------------------------------------------------------------

	// types of VST op code groups
	enum vstOpCodeGroup {
		vstFunctions,			// the C interface functions
		vstDispatcherOpCodes,	// plugin op codes
		vstAudioMasterOpCodes,	// host op codes
		vstNumOpCodeGroups
	};

	//------------------------------------------------------------------------

	// set as JUCE current logger and start logging, name is used until plug name is set
	void	init( const String& name );

	//set/get logged plugin name
	void setPlugName( const String& name ) { plugName = name; }
	const String getPlugName() { return plugName; }

	// load logger state from an XML element
	void loadState( const XmlElement& );

	// save logger state to an empty XML element, deleting all existing data in the element
	void saveState( XmlElement& );

#ifdef JUCE_WIN32
	// log DllMain entries
	void logDllMain( long dwReason );
#endif

	// stop/start logger (without changing enable state of opcodes)
	void stop() { logOn = false; }
	void start() { logOn = true; }

	// return true logging is on
	bool isOn() { return logOn; }

	// log functions parameters
	void setLogParameters( bool logParams_ ) { logParams = logParams_; }
	bool isLogParameters() { return logParams; }

	// log function return parameters
	void setLogReturnParameters( bool logRetParams_ ) { logRetParams = logRetParams_; }
	bool isLogReturnParameters() { return logRetParams; }

	//------------------------------------------------------------------------

	// more efficient log (not using String)
	virtual void log( const char* );

	// log the opcodes coming from the host
	void logVstDispatcher( long opCode, long index, long value, void *ptr, float opt );

	// log the return value returned by the dispatcher(if any, long and ptr)
	void logVstDispatcherReturn( long opCode, void *ptr, long result );

	// log the opcodes sent to the host
	void logVstAudioMaster( long opCode, long index, long value, void *ptr, float opt );

	// log the return value returned by the audio master(if any, long and ptr)
	void logVstAudioMasterReturn( long opCode, void *ptr, long result );

	// log process functions
	void logVstProcess( float **inputs, float **outputs, long sampleframes, bool replacing );

	// log set/get parameter functions
	void logVstSetParameter( long index, float parameter );
	void logVstGetParameter( long index );
	void logVstGetParameterReturn( float parameter );

	//------------------------------------------------------------------------

	// get name of opCode
	const String getVstOpCodeName( vstOpCodeGroup group, long opCode );

	// get number of opCodes for a given vst function
	int getVstNumOpCodes( vstOpCodeGroup group );

	// get dynamic debug info for vstOpCodeGroup and opCode
	long getVstOpCodeCount( vstOpCodeGroup group, long opCode );
	bool isVstOpCodeEnabled( vstOpCodeGroup group, long opCode );

	// set enabled state for vstOpCodeGroup and opCode
	void setVstOpCodeEnabled( bool enabled, vstOpCodeGroup group, long opCode );

	// set enabled state for all op codes
	void setVstOpCodesEnabled( bool enabled );

	// toggle enabled state for all op codes
	void toggleVstOpCodesEnabled();

	// reset all counters
	void resetVstOpCodeCounters();

	//------------------------------------------------------------------------
	
protected:

	// JUCE Logger override, used by writeToLog
	virtual void logMessage ( const String &message );

	String plugName;

	// true if log is started, false if stopped
	bool logOn;

	// true if logging functions parameters
	bool logParams;

	// true if logging functions return parameters
	bool logRetParams;

	//------------------------------------------------------------------------
};
